package Zim::Repository::Files;

use strict;
use POSIX qw(strftime);
use Encode;
use IO::File;
use File::Spec;
use File::Copy;
use File::MimeInfo;
use Zim::File;
use Zim::Repository::Base;
use Zim::Page::Text;

our $VERSION = '0.15';
our @ISA = qw/Zim::Repository::Base/;

my $case_tolerant_fs = File::Spec->case_tolerant();

=head1 NAME

Zim::Repository::Files - A file system based repository

=head1 DESCRIPTION

This module implements a file system based repository for zim.
See L<Zim::Repository> for the interface documentation.

=head1 METHODS

=over 4

=item new(NAMESPACE, DIR)

Simple constructor. DIR is the root directory of the repository.
NAMESPACE is the namespace that maps to that directory.

=cut

sub init { # called by new
	my ($self, $dir, $format) = @_;
	
	$dir = Zim::File->abs_path($dir);
	$self->{dir} = $dir;
	$self->{format} = $format || 'wiki';
	$self->{cache} = Zim::File->new($dir, '.zim.cache');
	$self->{cache_mtime} = (stat($self->{cache}))[9] || 0;
	#warn "Cache: $self->{cache}, last modified: $self->{cache_mtime}\n";
	
	$self->{ext} = ($self->{format} eq 'wiki') ? 'txt' : 'html';
		# FIXME HACK FIXME - this belongs in a Formats.pm
	
	return $self;
}

=item list_pages(NAMESPACE)

=cut

sub list_pages {
	my ($self, $namespace) = @_;
	
	my $dir = $self->dir($namespace);
	#print "list pages $namespace => $dir\n";
	return () unless -d $dir;

	my $mtime = (stat $dir)[9];
	return $self->_cache_dir($namespace, $dir)
		if $mtime > $self->{cache_mtime};
	
	#warn "Reading $namespace from cache\n";
	my @pages;
	for ($self->{cache}->read) {
		/^\s*\Q$namespace\E([^:\s]+:?)\s/ or next;
		push @pages, $1;
	}
	return sort {lc($a) cmp lc($b)} @pages;
}

sub _flush_cache {
	my $self = shift;
	$self->{cache}->remove;
	$self->{cache_mtime} = 0;
}

sub _cache_dir {
	my ($self, $namespace, $dir) = @_;
	warn "Indexing $namespace\n";
	
	my @pages = grep {defined $_} map {
		my $item = File::Spec->catfile($dir, $_);
		if (-d $item) { s/(_files)?$/:/ }
		else {
			if (mimetype($item) =~ /^text/) { s/\.$$self{ext}$// }
			else { $_ = undef }
		}
		$_;
	} Zim::File->list_dir($dir);
	
	@pages = sort {lc($a) cmp lc($b)} @pages;
	for (0 .. $#pages) { # cut doubles due to directories
		$pages[$_] = undef if $pages[$_+1] eq $pages[$_].':' ;
	}
	@pages = grep {defined $_} @pages;
	
	return @pages if $self->{parent}{config}{read_only};
	
	my $index = join '', grep {! /^\s*\Q$namespace\E([^:\s]+:?)\s/ }
		$self->{cache}->read;
	for my $p (@pages) { $index .= $self->_cache_string($namespace.$p) }
	$self->{cache}->write( $index );
	
	return @pages;
}

sub _cache_page {
	my ($self, $page) = @_;
	my $name = $page->name;
	my ($index, $is_dir);
	for ($self->{cache}->read) {
		if (/^\s*\Q$name\E(:?)\s/) { $is_dir = $1 }
		else { $index .= $_ }
	}
	$self->{cache}->write($index, $self->_cache_string($page, $is_dir))
}

sub _cache_string {
	my ($self, $page, $is_dir) = @_;
	unless (ref $page) {
		$is_dir = ($page =~ /:$/);
		$page = $self->get_page($page);
	}
	my @links;
	for (map $$_[1], $page->_list_links) {
		my ($t, $link) = $page->parse_link($_);
		next unless $t eq 'page';
		$link = $page->resolve_link($link);
		push @links, $link->name;
	}
	$self->wipe_array(\@links);
	my $key = $is_dir ? $page->name.':' : $page->name;
	return $key . ' => ' . join(' ', @links) . "\n" ;
}

sub _list_backlinks {
	my ($self, $page) = @_;
	my $name = $page->name;
	my @links;
	for ($self->{cache}->read) {
		/^\s*(\S+?):?\s+\=\>.*?\s\Q$name\E\s/ or next;
		push @links, $1;
	}
	return @links;
}

sub _search {
	my ($self, $query) = @_;
	
	my @matches;
	for ($self->{cache}->read) {
		/^\s*(\S+)\s+/ or next;
		my $match = $self->file($1)->grep(qr/(?i)\Q$query\E/);
		push @matches, [$1, $match] if $match;
		#warn "$1 matches $match times\n" if $match;
	}
	
	return @matches;
}

=item get_page(PAGE_NAME)

Returns an object of the type L<Zim::Page::Text>.

=cut

sub get_page {
	my ($self, $name, $source) = @_; # source is a private argument

	my $page = Zim::Page::Text->new($self, $name);
	$source ||= $self->file($name); # case sensitive lookup
	$page->set_source($source);
	$page->set_format($self->{format});
	$page->properties->{base} = $source->dir;
	
	unless ($source->exists) {
		$page->{parse_tree} = $self->_template($page);
		$page->status('new');
	}

	return $page;
}

=item C<resolve_page(NAME)>

Like C<get_page()> but with case in-sensitive lookup for a page name.

=cut

sub resolve_page {
	my ($self, $name) = @_;
	$name =~ s/[^\w\.\:\-]/_/g;
	my $source = $self->file($name, 1); # case tolerant lookup
	$name = $self->pagename($source->path);
	#warn "Resolved page: $name => $file => $page\n";
	$self->get_page($name, $source);
}

sub _template {
	# FIXME make template configurable
	my ($self, $page) = @_;
	$page->name =~ /([^:]+):*$/;
	my $title = ucfirst($1);
	$title =~ s/_/ /g;
	my $date = Encode::decode($Zim::CODESET,
		strftime('%A %d/%m/%Y %H:%M', localtime)   );
	return	['Document', {},
			['head1', {}, $title],
			['Para',  {empty_lines => 1}, "Created $date\n"]
		];
}

=item copy_page(SOURCE, TARGET)

=cut

sub copy_page {
	my ($self, $old, $new) = @_;
	my $source = $self->file($old);
	my $target = $self->file($new);
	Zim::File->copy($source, $target);
	@$new{'status', 'parse_tree'} = ('', undef);
}

=item move_page(SOURCE, TARGET)

=cut

sub move_page {
	my ($self, $old, $new) = @_;
	
	# Move file
	my $source = $self->file($old);
	my $target = $self->file($new);
	Zim::File->move($source, $target) if $source->exists;

	# Move tree below file
	my $old_tree = $self->dir($old);
	my $new_tree = $self->dir($new);
	Zim::File->move($old_tree, $new_tree) if -d $old_tree;
	
	# update objects
	@$old{'status', 'parse_tree'} = ('deleted', undef) if ref $old;
	@$new{'status', 'parse_tree'} = ('', undef)        if ref $new;
}

=item delete_page(PAGE)

=cut

sub delete_page {
	my ($self, $page) = @_;

	my $file = $self->file($page);
	my $dir = $file->dir;
	if ($file->exists) { $file->remove }
	else { # border case where empty dir was left for some reason
		$dir = $self->dir($page);
		Zim::File->remove_dir($dir);
	}
	
	@$page{'status', 'parse_tree'} = ('deleted', undef) if ref $page;
}

=item C<search()>

=cut

sub search {
	my ($self, $page, $query) = @_;
	
}

=back

=head2 Private methods

=over 4

=item C<file(PAGE, CASE)>

Returns a L<Zim::File> object for a page name.

CASE is a boolean that triggers a case in-sensitive lookup when true.

=item C<dir(PAGE, CASE)>

Returns a dir for a page name. This dir maps to the namespace below this page.

CASE is a boolean that triggers a case in-sensitive lookup when true.

=cut

sub dir {
	my ($self, $page, $case_tolerant) = @_;
	$self->file($page, $case_tolerant, 'DIR');
}

sub file {
	# TODO optimise using File::Glob
	# TODO make file() call dir() instead of the other way
	my ($self, $page, $case_tolerant, $is_dir) = @_;

	if (ref $page) { # page object
		return $page->{source} if defined $page->{source} and ! $is_dir;
		$page = $page->name;
	}
	
	$page =~ s/^:*$self->{namespace}:*//i;
	$page =~ s/:+$//;
	
	my @dirs = split /:+/, $page;
	my $basename = pop @dirs unless $is_dir;

	# Find the right dir
	my $dir = File::Spec->catdir($self->{dir}, @dirs);
	unless (!$case_tolerant_fs and  -d $dir) {
		$dir = $self->{dir};
		for (0 .. $#dirs) {
			my $new_dir = $self->_grep_dir(
				$dir, $dirs[$_], 1, $case_tolerant );
			if (defined $new_dir) { $dir = $new_dir }
			else {
				$dir = File::Spec->catdir($dir, @dirs[$_ .. $#dirs]);
				last;
			}
		}
	}
	
	return $dir if $is_dir;

	# Check the filename
	my $file = $self->_grep_dir($dir, $basename, 0, $case_tolerant);
	$file = File::Spec->catfile($dir, $basename .'.'.$$self{ext})
		unless defined $file;

	#warn "resolved $page to file $file\n";
	return Zim::File->new($file);
}

sub _grep_dir { # find a file or subdir
	my ($self, $dir, $basename, $is_dir, $case_tolerant) = @_;
	return undef unless -d $dir;
	my $extended = $basename . ($is_dir ? '_files' : '.'.$$self{ext});
	my $wrong_type = 0;
	
	unless ($case_tolerant and $case_tolerant_fs) {
		if ($is_dir) {
			my $path = File::Spec->catdir($dir, $basename);
			return $path if -d $path;
			$wrong_type = 1 if -e _;
		}
		else {
			my $path = File::Spec->catfile($dir, $basename);
			return $path if -f $path;
			$wrong_type = 1 if -e _;
		}
	
		my $path = $is_dir
			? File::Spec->catdir($dir, $extended)
			: File::Spec->catfile($dir, $extended) ;
		return $path if $wrong_type or -e $path;
		return undef unless $case_tolerant;
	}
	
	my @check = (lc($basename), lc($extended));
	my $suggestion;
	my $other_type = lc $basename . ($is_dir ? '.'.$$self{ext} : '_files');
	opendir DIR, $dir or die "Could not list dir $dir\n";
	while (my $item = readdir DIR) { 
		# FIXME for case_tolerant_fs it would be nicer to check
		# case sensitive version first in the listing
		eval { $item = Encode::decode_utf8($item, 1) };
		if (grep {$_ eq lc($item)} @check) {
			if ($is_dir) {
				$item = File::Spec->catdir($dir, $item);
				unless (-d $item) {
					$suggestion = $item;
					next;
				}
			}
			else {
				$item = File::Spec->catfile($dir, $item);
				unless (-f $item) {
					$suggestion = $item;
					next;
				}
			}
			closedir DIR;
			return $item;
		}
		elsif (lc($item) eq $other_type) {
			$suggestion = $is_dir
				? File::Spec->catdir($dir, $item)
		       		: File::Spec->catfile($dir, $item) ;
		}
	}
	closedir DIR;

	if ($suggestion) {
		if ($is_dir) { $suggestion =~ s/\.$$self{ext}$//          }
		else         { $suggestion =~ s/(_files)?$/\.$$self{ext}/ }
		return $suggestion;
	}
	
	return undef;
}

=item pagename(FILE)

Returns the page name corresponding to FILE. FILE does not actually
need to exist and can be a directory as well as a file.

=cut

sub pagename {
	my ($self, $file) = @_;
	$file = File::Spec->abs2rel($file, $self->{dir})
		if File::Spec->file_name_is_absolute($file);
	my @parts = map {s/_files$//; $_} grep length($_), File::Spec->splitdir($file);
	$parts[-1] =~ s/\.$$self{ext}$//;
	return $self->{namespace} . join ':', @parts;
}

=item C<get_source(SOURCE, MODE)>

Returns an IO::File object.

=cut

sub get_source {
	my ($self, $source, $mode) = @_;
	die "unknown mode: $mode" unless $mode eq 'r' or $mode eq 'w';
	$source->make_dir if $mode eq 'w';
	return $source->open($mode);
}

# Methods below for filesystem interaction - subtle differences with File::Path

sub _mkdir {
	my ($dir, $file) = @_; # if $file is true $dir is a file
	my ($vol, $dirs) = File::Spec->splitpath($dir, $file ? 0 : 1);
	my @dirs = File::Spec->splitdir($dirs);
	my $path = File::Spec->catpath($vol, shift @dirs);
	mkdir $path or die "Could not create dir $path\n"
		if length $path and ! -d $path;
	while (@dirs) {
		$path = File::Spec->catdir($path, shift @dirs);
		mkdir $path or die "Could not create dir $path\n"
			unless -d $path;
	}
}

sub _rmdir {
	my $dir = shift;
	#print STDERR "rmdir $dir\n";
	rmdir $dir or return; # fails when non-empty
	my ($vol, $dirs) = File::Spec->splitpath($dir, 1);
	my @dirs = File::Spec->splitdir($dirs);
	while (@dirs) {
		pop @dirs;
		$dir = File::Spec->catdir($vol, @dirs);
		#print STDERR "rmdir $dir\n";
		rmdir $dir or last; # fails when non-empty
	}
}

1;

__END__

=back

=head1 BUGS

Please mail the author if you find any bugs.

=head1 AUTHOR

Jaap Karssenberg || Pardus [Larus] E<lt>pardus@cpan.orgE<gt>

Copyright (c) 2005 Jaap G Karssenberg. All rights reserved.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zim>, L<Zim::Page>

=cut
